@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
  *  Copyright (C) 2004,2005 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -41,6 +40,7 @@ extern int verbose;
 
 
 struct aptmethod {
+	/*@only@*/ /*@null@*/
 	struct aptmethod *next;
 	char *name;
 	char *baseuri;
@@ -99,27 +99,23 @@ static void aptmethod_free(/*@only@*/struct aptmethod *method) {
 }
 
 retvalue aptmethod_shutdown(struct aptmethodrun *run) {
-	struct aptmethod *method,*lastmethod;
+	struct aptmethod *method,*lastmethod,**method_ptr;
 
 	/* first get rid of everything not running: */
-	method = run->methods; lastmethod = NULL;
-	while( method != NULL ) {
-		struct aptmethod *h;
+	method_ptr = &run->methods;
+	while( *method_ptr != NULL ) {
 
-		if( method->child > 0 ) {
+		if( (*method_ptr)->child > 0 ) {
 			if( verbose > 5 )
-				fprintf(stderr,"Still waiting for %d\n",(int)method->child);
-			lastmethod = method;
-			method = method->next;
+				fprintf(stderr,"Still waiting for %d\n",(int)(*method_ptr)->child);
+			method_ptr = &(*method_ptr)->next;
 			continue;
 		} else {
-			h = method->next;
-			if( lastmethod != NULL )
-				lastmethod->next = h;
-			else
-				run->methods = h;
-			aptmethod_free(method);
-			method = h;
+			/*@only@*/ struct aptmethod *h;
+			h = (*method_ptr);
+			*method_ptr = h->next;
+			h->next = NULL;
+			aptmethod_free(h);
 		}
 	}
 
@@ -289,6 +285,11 @@ inline static retvalue aptmethod_startup(struct aptmethodrun *run,struct aptmeth
 		return RET_ERRNO(err);
 	}
 
+	if( interupted() ) {
+		(void)close(stdin[0]);(void)close(stdin[1]);
+		(void)close(stdout[0]);(void)close(stdout[1]);
+		return RET_ERROR_INTERUPTED;
+	}
 	f = fork();
 	if( f < 0 ) {
 		int err = errno;
@@ -952,6 +953,9 @@ static retvalue senddata(struct aptmethod *method) {
 			return RET_OK;
 		}
 
+		if( interupted() )
+			return RET_ERROR_INTERUPTED;
+
 		method->alreadywritten = 0;
 		// TODO: make sure this is already checked for earlier...
 		assert(strchr(method->nexttosend->uri,'\n')==NULL || strchr(method->nexttosend->filename,'\n') == 0);
@@ -1078,6 +1082,7 @@ static retvalue readwrite(struct aptmethodrun *run,int *workleft,filesdb filesdb
 	v = select(maxfd+1,&readfds,&writefds,NULL,NULL);
 	if( v < 0 ) {
 		int err = errno;
+		//TODO: handle (err == EINTR) && interupted() specially
 		fprintf(stderr,"Select returned error: %d=%m\n",err);
 		*workleft = -1;
 		// TODO: what to do here?
@@ -1088,11 +1093,11 @@ static retvalue readwrite(struct aptmethodrun *run,int *workleft,filesdb filesdb
 
 	maxfd = 0;
 	for( method = run->methods ; method != NULL ; method = method->next ) {
-		if( FD_ISSET(method->stdout,&readfds) ) {
+		if( method->stdout != -1 && FD_ISSET(method->stdout,&readfds) ) {
 			r = receivedata(method,filesdb);
 			RET_UPDATE(result,r);
 		}
-		if( FD_ISSET(method->stdin,&writefds) ) {
+		if( method->stdin != -1 && FD_ISSET(method->stdin,&writefds) ) {
 			r = senddata(method);
 			RET_UPDATE(result,r);
 		}
@@ -1122,6 +1127,7 @@ retvalue aptmethod_download(struct aptmethodrun *run,const char *methoddir,files
 	  RET_UPDATE(result,r);
 	  r = readwrite(run,&workleft,filesdb);
 	  RET_UPDATE(result,r);
+	  // TODO: check interupted here...
 	} while( workleft > 0 );
 
 	return result;
