@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2003,2004,2005 Bernhard R. Link
+ *  Copyright (C) 2003,2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -31,6 +30,7 @@
 #include "ignore.h"
 #include "strlist.h"
 #include "md5sum.h"
+#include "copyfile.h"
 #include "names.h"
 #include "dirs.h"
 #include "chunks.h"
@@ -87,6 +87,9 @@ struct fileentry {
 	char *component;
 	/* only set after changes_includefiles */
 	char *filekey;
+	/* was already found in the pool before */
+	bool_t wasalreadythere;
+	bool_t included;
 	/* set between checkpkg and includepkg */
 	union { struct dscpackage *dsc; struct debpackage *deb;} pkg;
 };
@@ -108,6 +111,8 @@ struct changes {
 	char *srcdirectory;
 	/* (only to warn if multiple are used) */
 	const char *firstcomponent;
+	/* the directory the .changes file resides in */
+	char *incomingdirectory;
 };
 
 static void freeentries(/*@only@*/struct fileentry *entry) {
@@ -144,6 +149,7 @@ static void changes_free(/*@only@*/struct changes *changes) {
 		free(changes->srcdirectory);
 		free(changes->changesfilekey);
 //		trackedpackage_free(changes->trackedpkg);
+		free(changes->incomingdirectory);
 	}
 	free(changes);
 }
@@ -266,6 +272,12 @@ static retvalue newentry(struct fileentry **entry,const char *fileline,const cha
 			type = fe_DIFF;
 		else if( p-versionstart > 4 && strncmp(p-4,".dsc",4) == 0 )
 			type = fe_DSC;
+		else if( p-versionstart > 13 && strncmp(p-13,".orig.tar.bz2",13) == 0 )
+			type = fe_ORIG;
+		else if( p-versionstart > 8 && strncmp(p-8,".tar.bz2",8) == 0 )
+			type = fe_TAR;
+		else if( p-versionstart > 9 && strncmp(p-9,".diff.bz2",9) == 0 )
+			type = fe_DIFF;
 		else {
 			type = fe_UNKNOWN;
 			fprintf(stderr,"Unknown filetype: '%s', assuming to be source format...\n",fileline);
@@ -417,10 +429,12 @@ static retvalue changes_read(const char *filename,/*@out@*/struct changes **chan
 	r = check(filename,c,"Changes");
 	R;
 	r = chunk_getextralinelist(c->control,"Files",&filelines);
-	E("Missing 'Files' field");
+	E("Missing 'Files' field!");
 	r = changes_parsefilelines(filename,c,&filelines,packagetypeonly,forcearchitecture);
 	strlist_done(&filelines);
 	R;
+	r = dirs_getdirectory(filename,&c->incomingdirectory);
+	R;
 
 	*changes = c;
 	return RET_OK;
@@ -572,7 +586,7 @@ static inline retvalue checkforarchitecture(const struct fileentry *e,const char
 	return RET_OK;
 }
 
-static retvalue changes_check(const char *filename,struct changes *changes,/*@null@*/const char *forcearchitecture) {
+static retvalue changes_check(const char *filename,struct changes *changes,/*@null@*/const char *forcearchitecture, /*@null@*/const char *packagetypeonly) {
 	int i;
 	struct fileentry *e;
 	retvalue r = RET_OK;
@@ -591,13 +605,30 @@ static retvalue changes_check(const char *filename,struct changes *changes,/*@nu
 		r = checkforarchitecture(changes->files,forcearchitecture);
 		if( RET_WAS_ERROR(r) )
 			return r;
-	} else {
+	} else if( packagetypeonly == NULL ) {
 		for( i = 0 ; i < changes->architectures.count ; i++ ) {
+			const char *architecture = changes->architectures.values[i];
 			r = checkforarchitecture(changes->files,
-				changes->architectures.values[i]);
+					architecture);
 			if( RET_WAS_ERROR(r) )
 				return r;
 		}
+	} else if( strcmp(packagetypeonly,"dsc") == 0 ) {
+		if( strlist_in(&changes->architectures,"source") ) {
+			r = checkforarchitecture(changes->files, "source");
+			if( RET_WAS_ERROR(r) )
+				return r;
+		}
+	} else {
+		for( i = 0 ; i < changes->architectures.count ; i++ ) {
+			const char *architecture = changes->architectures.values[i];
+			if( strcmp(architecture,"source") != 0 ) {
+				r = checkforarchitecture(changes->files,
+						architecture);
+				if( RET_WAS_ERROR(r) )
+					return r;
+			}
+		}
 	}
 	/* Then check for each file, if its architecture is sensible
 	 * and listed. */
@@ -605,7 +636,7 @@ static retvalue changes_check(const char *filename,struct changes *changes,/*@nu
 	while( e != NULL ) {
 		if( !strlist_in(&changes->architectures,e->architecture) ) {
 			if( !IGNORING_(surprisingarch,
-			"'%s' looks like architecture '%s', but this is not listed in the Architecture-Header!\n",filename,e->architecture))
+			"'%s' looks like architecture '%s', but this is not listed in the Architecture-Header!\n",e->basename,e->architecture))
 				r = RET_ERROR;
 		}
 		if( e->type == fe_DSC ) {
@@ -655,7 +686,12 @@ static retvalue changes_check(const char *filename,struct changes *changes,/*@nu
 		fprintf(stderr,"I don't know what to do having a .tar.gz not being a .orig.tar.gz and a .diff.gz in '%s'!\n",filename);
 		return RET_ERROR;
 	}
-	if( strlist_in(&changes->architectures,"source") && !havedsc ) {
+	if( strlist_in(&changes->architectures,"source") && !havedsc &&
+			( forcearchitecture == NULL 
+			  || strcmp(forcearchitecture,"source") == 0 ) &&
+			( packagetypeonly == NULL 
+			  || strcmp(packagetypeonly,"dsc") == 0 )
+			) {
 		fprintf(stderr,"I don't know what to do with a source-upload not containing a .dsc in '%s'!\n",filename);
 		return RET_ERROR;
 	}
@@ -667,19 +703,14 @@ static retvalue changes_check(const char *filename,struct changes *changes,/*@nu
 	return r;
 }
 
-static retvalue changes_includefiles(filesdb filesdb,const char *filename,struct changes *changes,int delete) {
+static retvalue changes_checkfiles(filesdb filesdb,const char *filename,struct changes *changes) {
 	struct fileentry *e;
 	retvalue r;
-	char *sourcedir; 
-
-	r = dirs_getdirectory(filename,&sourcedir);
-	if( RET_WAS_ERROR(r) )
-		return r;
 
 	r = RET_NOTHING;
 
-	e = changes->files;
-	while( e != NULL ) {
+	for( e = changes->files; e != NULL ; e = e->next ) {
+		//TODO: decide earlier which files to include
 		if( FE_SOURCE(e->type) ) {
 			assert(changes->srcdirectory!=NULL);
 			e->filekey = calc_dirconcat(changes->srcdirectory,e->basename);
@@ -692,26 +723,113 @@ static retvalue changes_includefiles(filesdb filesdb,const char *filename,struct
 			directory = calc_sourcedir(e->component,changes->source);
 			if( directory == NULL )
 				return RET_ERROR_OOM;
-
 			e->filekey = calc_dirconcat(directory,e->basename);
 			free(directory);
 		}
 
-		if( e->filekey == NULL ) {
-			free(sourcedir);
+		if( e->filekey == NULL )
 			return RET_ERROR_OOM;
-		}
-		r = files_includefile(filesdb,sourcedir,e->basename,e->filekey,e->md5sum,NULL,delete);
+		/* do not copy yet, but only check if it could be included */
+		r = files_expect(filesdb,e->filekey,e->md5sum);
 		if( RET_WAS_ERROR(r) )
-			break;
-		e = e->next;
+			return r;
+		/* If is was already there, remember that */
+		if( RET_IS_OK(r) ) {
+			e->wasalreadythere = TRUE;
+		} else {
+		/* and if it needs inclusion check if there is a file */
+			char *fullfilename;
+
+			assert( r == RET_NOTHING );
+			// TODO: add a --paranoid to also check md5sums before copying?
+
+			fullfilename = calc_dirconcat(changes->incomingdirectory,e->basename);
+			if( fullfilename == NULL )
+				return RET_ERROR_OOM;
+			if( !isregularfile(fullfilename) ) {
+				fprintf(stderr, "Cannot find file '%s' needed by '%s'!\n", fullfilename,filename);
+				free(fullfilename);
+				return RET_ERROR_MISSING;
+			}
+			free(fullfilename);
+		}
 	}
 
-	free(sourcedir);
-	return r;
+	return RET_OK;
 }
 
-static retvalue changes_checkpkgs(filesdb filesdb,struct distribution *distribution,struct changes *changes,const struct alloverrides *ao, bool_t onlysigned) {
+static retvalue changes_includefiles(filesdb filesdb,struct changes *changes,int delete) {
+	struct fileentry *e;
+	retvalue r;
+
+	r = RET_NOTHING;
+
+	for( e = changes->files; e != NULL ; e = e->next ) {
+		assert( e->filekey != NULL );
+
+		if( e->wasalreadythere )
+			continue;
+
+		r = files_includefile(filesdb,changes->incomingdirectory,
+				e->basename,e->filekey,e->md5sum,NULL,
+				/* do not delete, we do that later outself */
+				(delete>=D_MOVE)?D_COPY:delete);
+		if( RET_IS_OK(r) )
+			e->included = TRUE;
+		if( RET_WAS_ERROR(r) )
+			return r;
+	}
+
+	return r;
+}
+/* run if packages are not all includeable and the stuff put into the
+ * pool shall be removed again */
+static void changes_unincludefiles(filesdb filesdb,struct changes *changes) {
+	struct fileentry *e;
+
+	for( e = changes->files; e != NULL ; e = e->next ) {
+
+		if( e->filekey == NULL || e->wasalreadythere || !e->included )
+			continue;
+
+		(void)files_deleteandremove(filesdb,e->filekey,TRUE,FALSE);
+	}
+}
+/* delete the files included */
+static retvalue changes_deleteleftoverfiles(struct changes *changes,int delete) {
+	struct fileentry *e;
+	retvalue result,r;
+
+	if( delete < D_MOVE )
+		return RET_OK;
+
+	result = RET_OK;
+	// TODO: we currently only see files included here, so D_DELETE
+	// only affacts the .changes file.
+
+	for( e = changes->files; e != NULL ; e = e->next ) {
+		char *fullorigfilename;
+
+		if( delete < D_DELETE && e->filekey == NULL )
+			continue;
+
+		fullorigfilename = calc_dirconcat(changes->incomingdirectory,
+					e->basename);
+
+		if( unlink(fullorigfilename) != 0 ) {
+			int e = errno;
+			fprintf(stderr, "Error deleting '%s': %d=%s\n",
+					fullorigfilename, e, strerror(e));
+			r = RET_ERRNO(e);
+			RET_UPDATE(result,r);
+		} 
+		free(fullorigfilename);
+	}
+
+	return result;
+}
+
+static retvalue changes_checkpkgs(filesdb filesdb,struct distribution *distribution,struct changes *changes,const struct alloverrides *ao, bool_t onlysigned, const char *sourcedirectory) {
 	struct fileentry *e;
 	retvalue r;
 	bool_t somethingwasmissed = FALSE;
@@ -753,7 +871,7 @@ static retvalue changes_checkpkgs(filesdb filesdb,struct distribution *distribut
 			assert(changes->srcdirectory!=NULL);
 			r = dsc_prepare(&e->pkg.dsc,filesdb,
 				changes->srccomponent,e->section,e->priority,
-				distribution,fullfilename,
+				distribution,sourcedirectory,fullfilename,
 				e->filekey,e->basename,
 				changes->srcdirectory,e->md5sum,
 				ao->dsc,D_INPLACE,onlysigned);
@@ -785,6 +903,9 @@ static retvalue changes_includepkgs(const char *dbdir,references refs,struct dis
 			e = e->next;
 			continue;
 		}
+		if( interupted() ) {
+			return RET_ERROR_INTERUPTED;
+		}
 		if( e->type == fe_DEB ) {
 			r = deb_addprepared(e->pkg.deb,dbdir,refs,
 				e->architecture,"deb",
@@ -822,16 +943,29 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 	retvalue result,r;
 	struct changes *changes;
 	struct trackingdata trackingdata;
+	char *directory;
 
 	r = changes_read(changesfilename,&changes,packagetypeonly,forcearchitecture,onlysigned);
 	if( RET_WAS_ERROR(r) )
 		return r;
 
+	if( IGNORABLE(missingfile) ) {
+		r = dirs_getdirectory(changesfilename,&directory);
+		if( RET_WAS_ERROR(r) ) {
+			changes_free(changes);
+			return r;
+		}
+	} else
+		directory = NULL;
+
 	if( (distribution->suite == NULL || 
 		!strlist_in(&changes->distributions,distribution->suite)) &&
 	    !strlist_in(&changes->distributions,distribution->codename) ) {
-		if( !IGNORING("Ignoring","To ignore",wrongdistribution,".changes put in a distribution not listed within it!\n") )
+		if( !IGNORING("Ignoring","To ignore",wrongdistribution,".changes put in a distribution not listed within it!\n") ) {
+			free(directory);
+			changes_free(changes);
 			return RET_ERROR;
+		}
 	}
 
 	/* look for component, section and priority to be correct or guess them*/
@@ -839,18 +973,28 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 
 	/* do some tests if values are sensible */
 	if( !RET_WAS_ERROR(r) )
-		r = changes_check(changesfilename,changes,forcearchitecture);
+		r = changes_check(changesfilename,changes,forcearchitecture,packagetypeonly);
+
+	if( interupted() )
+		RET_UPDATE(r,RET_ERROR_INTERUPTED);
+
+	if( !RET_WAS_ERROR(r) )
+		r = changes_checkfiles(filesdb,changesfilename,changes);
+
+	if( interupted() )
+		RET_UPDATE(r,RET_ERROR_INTERUPTED);
 
 	/* add files in the pool */
-	//TODO: D_DELETE would fail here, what to do?
 	if( !RET_WAS_ERROR(r) )
-		r = changes_includefiles(filesdb,changesfilename,changes,delete);
+		r = changes_includefiles(filesdb,changes,delete);
 
 	if( !RET_WAS_ERROR(r) )
-		r = changes_checkpkgs(filesdb,distribution,changes,ao,onlysigned);
+		r = changes_checkpkgs(filesdb,distribution,changes,ao,onlysigned,directory);
 
+	free(directory);
 
 	if( RET_WAS_ERROR(r) ) {
+		changes_unincludefiles(filesdb,changes);
 		changes_free(changes);
 		return r;
 	}
@@ -873,8 +1017,11 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 				trackingdata_done(&trackingdata);
 				return RET_ERROR_OOM;
 			}
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
 			/* always D_COPY, and only delete it afterwards... */
-			r = files_include(filesdb,changesfilename,
+				r = files_include(filesdb,changesfilename,
 					changes->changesfilekey,
 					NULL,NULL,D_COPY);
 			if( RET_WAS_ERROR(r) ) {
@@ -884,33 +1031,39 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 			}
 		}
 	}
+	if( interupted() ) {
+		trackingdata_done(&trackingdata);
+		changes_free(changes);
+		return RET_ERROR_INTERUPTED;
+	}
 
 	/* add the source and binary packages in the given distribution */
 	result = changes_includepkgs(dbdir,refs,
 		distribution,changes,dereferencedfilekeys,
 		(tracks!=NULL)?&trackingdata:NULL);
 
-	if( RET_WAS_ERROR(r) ) {
+	if( RET_WAS_ERROR(result) ) {
 		if( tracks != NULL ) {
 			trackingdata_done(&trackingdata);
 		}
 		changes_free(changes);
-		return r;
+		return result;
 	}
 
 	if( tracks != NULL ) {
 		if( changes->changesfilekey != NULL ) {
+			char *changesfilekey = strdup(changes->changesfilekey);
 			assert( changes->srcdirectory != NULL );
+			if( changesfilekey == NULL ) {
+				trackingdata_done(&trackingdata);
+				changes_free(changes);
+				return RET_ERROR_OOM;
+			}
 
-			r = trackedpackage_addfilekey(tracks,trackingdata.pkg,ft_CHANGES,changes->changesfilekey,FALSE,refs);
+			r = trackedpackage_addfilekey(tracks,trackingdata.pkg,ft_CHANGES,changesfilekey,FALSE,refs);
 			RET_ENDUPDATE(result,r);
 		}
-		if( trackingdata.isnew ) {
-			r = tracking_put(tracks,trackingdata.pkg);
-		} else {
-			r = tracking_replace(tracks,trackingdata.pkg);
-		}
-		trackingdata_done(&trackingdata);
+		r = trackingdata_finish(tracks, &trackingdata, refs, dereferencedfilekeys);
 		RET_ENDUPDATE(result,r);
 		if( RET_WAS_ERROR(result) ) {
 			changes_free(changes);
@@ -918,9 +1071,9 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 		}
 	}
 
-	if( delete >= D_MOVE && changesfilename != NULL ) {
-		if( result == RET_NOTHING && delete < D_DELETE && 
-				changes->changesfilekey == NULL) {
+	if( (delete >= D_MOVE && changes->changesfilekey != NULL) ||
+			delete >= D_DELETE ) {
+		if( result == RET_NOTHING && delete < D_DELETE ) {
 			if( verbose >= 0 ) {
 				fprintf(stderr,"Not deleting '%s' as no package was added or some package was missed.\n(Use --delete --delete to delete anyway in such cases)\n",changesfilename);
 			}
@@ -933,7 +1086,8 @@ retvalue changes_add(const char *dbdir,trackingdb const tracks,references refs,f
 			}
 		}
 	}
+	result = changes_deleteleftoverfiles(changes, delete);
 	(void)changes_free(changes);
 
-	return RET_OK;
+	return result;
 }
