@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2003,2004,2005 Bernhard R. Link
+ *  Copyright (C) 2003,2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -37,7 +36,7 @@
 #include "reference.h"
 #include "binaries.h"
 #include "files.h"
-#include "extractcontrol.h"
+#include "debfile.h"
 #include "guesscomponent.h"
 #include "tracking.h"
 
@@ -307,8 +306,6 @@ retvalue deb_prepare(/*@out@*/struct debpackage **deb,filesdb filesdb,const char
 	assert( (givenmd5sum!=NULL && givenfilekey!=NULL ) ||
 		(givenmd5sum==NULL && givenfilekey==NULL ) );
 
-	//TODO: add some check here to make sure it is really a .deb file...
-
 	/* First taking a closer look to the file: */
 
 	r = deb_read(&pkg,debfilename,needsourceversion);
@@ -441,7 +438,7 @@ retvalue deb_addprepared(const struct debpackage *pkg, const char *dbdir,referen
 	retvalue r,result;
 	int i;
 
-	/* finaly put it into one or more architectures of the distribution */
+	/* finally put it into one or more architectures of the distribution */
 
 	result = RET_NOTHING;
 
@@ -450,7 +447,10 @@ retvalue deb_addprepared(const struct debpackage *pkg, const char *dbdir,referen
 		r = target_initpackagesdb(t,dbdir);
 		if( !RET_WAS_ERROR(r) ) {
 			retvalue r2;
-			r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ARCH_BINARY);
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
+				r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ARCH_BINARY);
 			r2 = target_closepackagesdb(t);
 			RET_ENDUPDATE(r,r2);
 		}
@@ -460,7 +460,10 @@ retvalue deb_addprepared(const struct debpackage *pkg, const char *dbdir,referen
 		r = target_initpackagesdb(t,dbdir);
 		if( !RET_WAS_ERROR(r) ) {
 			retvalue r2;
-			r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ALL_BINARY);
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
+				r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ALL_BINARY);
 			r2 = target_closepackagesdb(t);
 			RET_ENDUPDATE(r,r2);
 		}
@@ -473,14 +476,16 @@ retvalue deb_addprepared(const struct debpackage *pkg, const char *dbdir,referen
 		r = target_initpackagesdb(t,dbdir);
 		if( !RET_WAS_ERROR(r) ) {
 			retvalue r2;
-			r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ALL_BINARY);
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
+				r = target_addpackage(t,refs,pkg->package,pkg->version,pkg->control,&pkg->filekeys,FALSE,dereferencedfilekeys,trackingdata,ft_ALL_BINARY);
 			r2 = target_closepackagesdb(t);
 			RET_ENDUPDATE(r,r2);
 		}
 		RET_UPDATE(result,r);
 	}
-
-//	deb_free(pkg);
+	RET_UPDATE(distribution->status, result);
 
 	return result;
 }
@@ -515,12 +520,7 @@ retvalue deb_add(const char *dbdir,references refs,filesdb filesdb,const char *f
 
 	if( tracks != NULL ) {
 		retvalue r2;
-		if( trackingdata.isnew ) {
-			r2 = tracking_put(tracks,trackingdata.pkg);
-		} else {
-			r2 = tracking_replace(tracks,trackingdata.pkg);
-		}
-		trackingdata_done(&trackingdata);
+		r2 = trackingdata_finish(tracks, &trackingdata, refs, dereferencedfilekeys);
 		RET_ENDUPDATE(r,r2);
 	}
 	
