@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
  *  Copyright (C) 2003,2004,2005 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -34,7 +33,7 @@ extern int verbose;
 // of only \n terminated oned. Though this has still to be tested properly...
 
 /* Call action for each chunk in <filename> */
-retvalue chunk_foreach(const char *filename,chunkaction action,void *data,int force,bool_t stopwhenok){
+retvalue chunk_foreach(const char *filename,chunkaction action,void *data,bool_t stopwhenok){
 	gzFile f;
 	retvalue result,ret;
 	char *chunk;
@@ -46,20 +45,23 @@ retvalue chunk_foreach(const char *filename,chunkaction action,void *data,int fo
 	}
 	result = RET_NOTHING;
 	while( (ret = chunk_read(f,&chunk)) == RET_OK ) {
-		ret = action(data,chunk);
-
-		// in this one a segfault was reported, where did
-		// it come from?
-		free(chunk);
-
-		if( RET_WAS_ERROR(ret) && force <= 0 ) {
-			if( verbose > 0 )
-				fprintf(stderr,"Stop reading further chunks from '%s' due to privious errors.\n",filename);
+		if( interupted() ) {
+			RET_UPDATE(result,RET_ERROR_INTERUPTED);
+			free(chunk);
 			break;
 		}
+		ret = action(data,chunk);
+
+		free(chunk);
+
+		if( RET_WAS_ERROR(ret) ) {
+			if( verbose > 0 )
+				fprintf(stderr,"Stop reading further chunks from '%s' due to previous errors.\n",filename);
+			break;
+		}
+		RET_UPDATE(result,ret);
 		if( stopwhenok && RET_IS_OK(ret) )
 			break;
-		RET_UPDATE(result,ret);
 	}
 	RET_UPDATE(result,ret);
 	//TODO: check result:
@@ -695,6 +697,7 @@ char *chunk_replacefield(const char *chunk,const char *fieldname,const char *dat
 
 retvalue chunk_checkfields(const char *chunk,const char * const *allowedfieldnames,bool_t commentsallowed) {
 	const char *c = chunk;
+	long long alreadyfound = 0;
 
 	assert( chunk != NULL );
 	assert( allowedfieldnames != NULL );
@@ -742,12 +745,21 @@ overlongcomments,"A comment is continued by a line starting with space in '%s'!\
 
 		} else {
 			const char * const *allowedfn = allowedfieldnames;
+			long long i = 1;
 
-			for( ; *allowedfn != NULL ; allowedfn++ ) {
+			for( ; *allowedfn != NULL ; allowedfn++,i = i << 1 ) {
 				const char *allowed = *allowedfn;
 				if( strncasecmp(allowed,c,nameend-c) == 0 
-				    && allowed[nameend-c] == '\0' )
+				    && allowed[nameend-c] == '\0' ) {
+					if( (i&alreadyfound) != 0 ) {
+						if( !IGNORING("Ignoring","To ignore this",
+doublefield, "Doubled field name: %s (only first will be used)\n",allowed) ) {
+							return RET_ERROR;
+						}
+					}
+					alreadyfound |= i;
 					break;
+				}
 			}
 			if( *allowedfn == NULL ) {
 				char *fieldname = strndup(c,nameend-c);
