@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
  *  Copyright (C) 2003 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -71,6 +70,9 @@ retvalue dirs_make_parent(const char *filename) {
 retvalue dirs_make_recursive(const char *directory) {
 	retvalue r,result;
 	
+	if( interupted() ) {
+		return RET_ERROR_INTERUPTED;
+	}
 	r = dirs_make_parent(directory);
 	result = dirs_check(directory);
 	RET_UPDATE(result,r);
