@@ -5,6 +5,8 @@
 #include "globals.h"
 #endif
 
+bool_t interupted(void);
+
 /* retvalue is simply an int.
  * just named to show it follows the given semantics */
 typedef int retvalue;
@@ -18,6 +20,7 @@ typedef int retvalue;
 #define RET_ERROR_GPGME (retvalue)-5
 #define RET_ERROR_BADSIG (retvalue)-6
 #define RET_ERROR_MISSING (retvalue)-7
+#define RET_ERROR_INTERUPTED (retvalue)-8
 
 #define RET_IS_OK(r) ((r) == RET_OK)
 #define RET_WAS_NO_ERROR(r) ((r) >= (retvalue)0)
