@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2003,2004,2005 Bernhard R. Link
+ *  Copyright (C) 2003,2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -35,9 +34,13 @@
 #include "names.h"
 #include "files.h"
 #include "copyfile.h"
+#include "ignore.h"
+#include "filelist.h"
+#include "debfile.h"
 
 struct s_filesdb {
 	DB *database;
+	DB *contents;
 	char *mirrordir;
 };
 
@@ -82,7 +85,8 @@ retvalue files_initialize(filesdb *fdb,const char *dbpath,const char *mirrordir)
 		free(db);
 		return RET_DBERR(dbret);
 	}
-	if ((dbret = db->database->open(db->database, filename, "md5sums", DB_BTREE, DB_CREATE, 0664)) != 0) {
+	dbret = DB_OPEN(db->database, filename, "md5sums", DB_BTREE,DB_CREATE);
+	if (dbret != 0) {
 		db->database->err(db->database, dbret, "%s", filename);
 		(void)db->database->close(db->database,0);
 		free(filename);
@@ -91,22 +95,51 @@ retvalue files_initialize(filesdb *fdb,const char *dbpath,const char *mirrordir)
 		return RET_DBERR(dbret);
 	}
 	free(filename);
+	filename = calc_dirconcat(dbpath,"contents.cache.db");
+	if( filename == NULL ) {
+		(void)db->database->close(db->database,0);
+		free(db->mirrordir);
+		free(db);
+		return RET_ERROR_OOM;
+	}
+	if ((dbret = db_create(&db->contents, NULL, 0)) != 0) {
+		fprintf(stderr, "db_create: %s\n", db_strerror(dbret));
+		(void)db->database->close(db->database,0);
+		free(filename);
+		free(db->mirrordir);
+		free(db);
+		return RET_DBERR(dbret);
+	}
+	dbret = DB_OPEN(db->contents,filename,"filelists",DB_BTREE,DB_CREATE);
+	if( dbret != 0 ) {
+		db->contents->err(db->contents, dbret, "%s", filename);
+		(void)db->contents->close(db->contents,0);
+		(void)db->database->close(db->database,0);
+		free(filename);
+		free(db->mirrordir);
+		free(db);
+		return RET_DBERR(dbret);
+	}
+	free(filename);
 	*fdb = db;
 	return RET_OK;
 }
 
 /* release the files-database initialized got be files_initialize */
 retvalue files_done(filesdb db) {
-	int dberr;
+	int dberr,dberr2;
 
 	assert( db != NULL);
 
 	free( db->mirrordir );
 	/* just in case we want something here later */
 	dberr = db->database->close(db->database,0);
+	dberr2 = db->contents->close(db->contents,0);
 	free(db);
 	if( dberr != 0 )
 		return RET_DBERR(dberr);
+	else if( dberr2 != 0 )
+		return RET_DBERR(dberr2);
 	else
 		return RET_OK;
 }
@@ -152,15 +185,29 @@ static retvalue files_get(filesdb db,const char *filekey,/*@out@*/char **md5sum)
 
 
 /* remove file's md5sum from database */
-retvalue files_remove(filesdb db,const char *filekey) {
+retvalue files_remove(filesdb db,const char *filekey, bool_t ignoremissing) {
 	int dbret;
 	DBT key;
 
+	if( db->contents != NULL ) {
+		SETDBT(key,filekey);
+		(void)db->contents->del(db->contents,
+				NULL, &key, 0);
+	}
 	SETDBT(key,filekey);
-	if ((dbret = db->database->del(db->database, NULL, &key, 0)) == 0) {
+	dbret = db->database->del(db->database, NULL, &key, 0);
+	if( dbret == 0 ) {
 		if( verbose > 6 )
 			printf("db: %s: file forgotten.\n", (const char *)key.data);
 		return RET_OK;
+	} else if( dbret == DB_NOTFOUND ) {
+		if( ignoremissing )
+			return RET_NOTHING;
+		else {
+			fprintf(stderr, "To be forgotten filekey '%s' was not known.\n",
+					filekey);
+			return RET_ERROR_MISSING;
+		}
 	} else {
 		db->database->err(db->database, dbret, "files.db:");
 		return RET_DBERR(dbret);
@@ -168,11 +215,13 @@ retvalue files_remove(filesdb db,const char *filekey) {
 }
 
 /* delete the file and remove its md5sum from database */
-retvalue files_deleteandremove(filesdb filesdb,const char *filekey,bool_t rmdirs) {
+retvalue files_deleteandremove(filesdb filesdb,const char *filekey,bool_t rmdirs, bool_t ignoreifnot) {
 	int err,en;
 	char *filename;
 	retvalue r;
 
+	if( interupted() )
+		return RET_ERROR_INTERUPTED;
 	if( verbose >= 0 )
 		printf("deleting and forgetting %s\n",filekey);
 	filename = calc_fullfilename(filesdb->mirrordir,filekey);
@@ -183,7 +232,8 @@ retvalue files_deleteandremove(filesdb filesdb,const char *filekey,bool_t rmdirs
 		en = errno;
 		r = RET_ERRNO(en);
 		if( errno == ENOENT ) {
-			fprintf(stderr,"%s not found, forgetting anyway\n",filename);
+			if( !ignoreifnot )
+				fprintf(stderr,"%s not found, forgetting anyway\n",filename);
 		} else {
 			fprintf(stderr,"error while unlinking %s: %m(%d)\n",filename,en);
 			free(filename);
@@ -226,8 +276,7 @@ retvalue files_deleteandremove(filesdb filesdb,const char *filekey,bool_t rmdirs
 
 	}
 	free(filename);
-	r = files_remove(filesdb,filekey);
-	return r;
+	return files_remove(filesdb, filekey, ignoreifnot);
 }
 
 static retvalue files_calcmd5(/*@out@*/char **md5sum,const char *filename) {
@@ -309,7 +358,7 @@ retvalue files_expectfiles(filesdb filesdb,const struct strlist *filekeys,const
 		if( r == RET_NOTHING ) {
 			/* File missing */
 			fprintf(stderr,"Missing file %s\n",filekey);
-			return RET_ERROR;
+			return RET_ERROR_MISSING;
 		}
 	}
 	return RET_OK;
@@ -388,7 +437,22 @@ retvalue files_foreach(filesdb db,per_file_action action,void *privdata) {
 	CLEARDBT(data);	
 	result = RET_NOTHING;
 	while( (dbret=cursor->c_get(cursor,&key,&data,DB_NEXT)) == 0 ) {
-		r = action(privdata,(const char*)key.data,(const char*)data.data);
+		size_t fk_len = key.size-1;
+		const char *filekey = key.data;
+		size_t md_len = data.size-1;
+		const char *md5sum = data.data;
+
+		if( filekey[fk_len] != '\0' ) {
+			fprintf(stderr, "Incoherent data in file database!\n");
+			cursor->c_close(cursor);
+			return RET_ERROR;
+		}
+		if( md5sum[md_len] != '\0' ) {
+			fprintf(stderr, "Incoherent data in file database!\n");
+			cursor->c_close(cursor);
+			return RET_ERROR;
+		}
+		r = action(privdata,filekey,md5sum);
 		RET_UPDATE(result,r);
 	}
 	if( dbret != DB_NOTFOUND ) {
@@ -521,7 +585,6 @@ retvalue files_include(filesdb db,const char *sourcefilename,const char *filekey
 	retvalue r;
 	char *md5sumfound;
 
-	assert( md5sum == NULL || calculatedmd5sum == NULL );
 	if( md5sum != NULL ) {
 		r = files_expect(db,filekey,md5sum);
 		if( RET_WAS_ERROR(r) )
@@ -530,7 +593,13 @@ retvalue files_include(filesdb db,const char *sourcefilename,const char *filekey
 			if( delete >= D_MOVE ) {
 				copyfile_delete(sourcefilename);
 			}
-			return RET_NOTHING;
+			if( calculatedmd5sum != NULL ) {
+				char *n = strdup(md5sum);
+				if( n == NULL )
+					return RET_ERROR_OOM;
+				*calculatedmd5sum = n;
+			}
+			return RET_OK;
 		}
 	} else {
 		char *md5indatabase,*md5offile;
@@ -544,7 +613,7 @@ retvalue files_include(filesdb db,const char *sourcefilename,const char *filekey
 					*calculatedmd5sum = md5indatabase;
 				else
 					free(md5indatabase);
-				return RET_NOTHING;
+				return RET_OK;
 			}
 			r = files_calcmd5(&md5offile,sourcefilename);
 			if( RET_WAS_ERROR(r) ) {
@@ -570,9 +639,19 @@ retvalue files_include(filesdb db,const char *sourcefilename,const char *filekey
 			}
 		}
 	}
-	if( delete == D_INPLACE ) {
+	if( sourcefilename == NULL ) {
 		fprintf(stderr,"Unable to find %s/%s!\n",db->mirrordir,filekey);
+		if( delete == D_INPLACE ) {
+			fprintf(stderr,"Perhaps you forgot to give dpkg-buildpackage the -sa option,\n or you cound try --ignore=missingfile\n");
+		}
 		return RET_ERROR_MISSING;
+	} if( delete == D_INPLACE ) {
+		if( IGNORING("Looking around if it is elsewhere", "To look around harder, ",missingfile,"Unable to find %s/%s!\n",db->mirrordir,filekey))
+			r = copyfile_copy(db->mirrordir,filekey,sourcefilename,md5sum,&md5sumfound);
+		else
+			r = RET_ERROR_MISSING;
+		if( RET_WAS_ERROR(r) )
+			return r;
 	} else if( delete == D_COPY ) {
 		r = copyfile_copy(db->mirrordir,filekey,sourcefilename,md5sum,&md5sumfound);
 		if( RET_WAS_ERROR(r) )
@@ -601,9 +680,14 @@ retvalue files_includefile(filesdb db,const char *sourcedir,const char *basename
 	char *sourcefilename;
 	retvalue r;
 
-	sourcefilename = calc_dirconcat(sourcedir,basename);
-	if( sourcefilename == NULL )
-		return RET_ERROR_OOM;
+	if( sourcedir == NULL ) {
+		assert(delete == D_INPLACE);
+		sourcefilename = NULL;
+	} else {
+		sourcefilename = calc_dirconcat(sourcedir,basename);
+		if( sourcefilename == NULL )
+			return RET_ERROR_OOM;
+	}
 	r = files_include(db,sourcefilename,filekey,md5sum,calculatedmd5sum,delete);
 	free(sourcefilename);
 	return r;
@@ -616,7 +700,8 @@ retvalue files_includefiles(filesdb db,const char *sourcedir,const struct strlis
 	retvalue result,r;
 	int i;
 
-	assert( sourcedir != NULL ); assert( basefilenames != NULL );
+	assert( sourcedir != NULL || delete == D_INPLACE );
+	assert( basefilenames != NULL );
 	assert( filekeys != NULL ); assert( md5sums != NULL );
 	assert( basefilenames->count == filekeys->count );
 	assert( filekeys->count == md5sums->count );
@@ -638,3 +723,169 @@ retvalue files_includefiles(filesdb db,const char *sourcedir,const struct strlis
 char *files_calcfullfilename(const filesdb filesdb,const char *filekey) {
 	return calc_dirconcat(filesdb->mirrordir,filekey);
 }
+
+static retvalue files_addfilelist(filesdb db,const char *filekey,const char *filelist) {
+	int dbret;
+	DBT key,data;
+	const char *e;
+
+	SETDBT(key,filekey);
+	memset(&data,0,sizeof(data));
+	e = filelist;
+	while( *e != '\0' ) {
+		while( *e != '\0' )
+			e++;
+		e++;
+	}
+	e++;
+	data.data = (void*)filelist;
+	data.size = e-filelist;
+	dbret = db->contents->put(db->contents, NULL, &key, &data, 0);
+	if( dbret == 0) {
+		return RET_OK;
+	} else {
+		db->contents->err(db->contents, dbret, "filelists.db:");
+		return RET_DBERR(dbret);
+	}
+}
+
+retvalue files_getfilelist(filesdb db,const char *filekey,const struct filelist_package *package, struct filelist_list *filelist) {
+	int dbret;
+	DBT key,data;
+
+	SETDBT(key,filekey);
+	CLEARDBT(data);
+
+	if( (dbret = db->contents->get(db->contents, NULL, &key, &data, 0)) == 0){
+		const char *p = data.data; size_t size = data.size;
+
+		while( size > 0 && *p != '\0' ) {
+			size_t len = strnlen(p,size);
+			if( p[len] != '\0' ) {
+				fprintf(stderr,"Corrupt filelist for %s in filelists.db!\n",
+						filekey);
+				return RET_ERROR;
+			}
+			filelist_add(filelist, package, p);
+			p += len+1;
+			size -= len+1;
+		}
+		if( size != 1 || *p != '\0' ) {
+			fprintf(stderr,"Corrupt filelist for %s in filelists.db!\n",
+					filekey);
+			return RET_ERROR;
+		}
+		return RET_OK;
+	} else if( dbret != DB_NOTFOUND ){
+		 db->contents->err(db->contents, dbret, "filelists.db:");
+		 return RET_DBERR(dbret);
+	}
+	return RET_NOTHING;
+}
+
+retvalue files_genfilelist(filesdb db,const char *filekey,const struct filelist_package *package, struct filelist_list *filelist) {
+	char *debfilename = calc_dirconcat(db->mirrordir, filekey);
+	char *contents;
+	const char *p;
+	retvalue result,r;
+
+	if( debfilename == NULL ) {
+		return RET_ERROR_OOM;
+	}
+	r = getfilelist(&contents, debfilename);
+	free(debfilename);
+	if( !RET_IS_OK(r) )
+		return r;
+	result = files_addfilelist(db, filekey, contents);
+	for( p = contents; *p != '\0'; p += strlen(p)+1 ) {
+		r = filelist_add(filelist, package, p);
+		RET_UPDATE(result,r);
+	}
+	free(contents);
+	return result;
+}
+
+retvalue files_regenerate_filelist(filesdb db, bool_t reread) {
+	DBC *cursor;
+	DBT key,data;
+	int dbret;
+	retvalue result,r;
+
+	assert( db->contents != NULL );
+
+	cursor = NULL;
+	if( (dbret = db->database->cursor(db->database,NULL,&cursor,0)) != 0 ) {
+		db->database->err(db->database, dbret, "files.db:");
+		return RET_DBERR(dbret);
+	}
+	CLEARDBT(key);	
+	CLEARDBT(data);	
+	result = RET_NOTHING;
+	while( (dbret=cursor->c_get(cursor,&key,&data,DB_NEXT)) == 0 ) {
+		size_t l = key.size-1;
+		const char *filekey = key.data;
+
+		if( filekey[l] != '\0' ) {
+			fprintf(stderr, "File database is in a broken shape!\n");
+			cursor->c_close(cursor);
+			return RET_ERROR;
+		}
+		if( l > 4 && strcmp(filekey+l-4,".deb") == 0 ) {
+			bool_t needed;
+
+			if( reread )
+				needed = TRUE;
+			else {
+				DBT listkey,listdata;
+
+				SETDBT(listkey,filekey);
+				CLEARDBT(listdata);
+
+				dbret = db->contents->get(db->contents, NULL, &listkey, &listdata, 0);
+				needed = dbret != 0;
+			}
+			if( needed ) {
+				char *debfilename;
+				char *filelist;
+
+				debfilename = calc_dirconcat(db->mirrordir, filekey);
+				if( debfilename == NULL ) {
+					cursor->c_close(cursor);
+					return RET_ERROR_OOM;
+				}
+
+				r = getfilelist(&filelist, debfilename);
+				free(debfilename);
+				if( RET_IS_OK(r) ) {
+					if( verbose > 0 )
+						puts(filekey);
+					if( verbose > 6 ) {
+						const char *p = filelist;
+						while( *p != '\0' ) {
+							putchar(' ');
+							puts(p);
+							p += strlen(p)+1;
+						}
+					}
+					r = files_addfilelist(db, filekey, filelist);
+					free(filelist);
+				}
+				if( RET_WAS_ERROR(r) ) {
+					cursor->c_close(cursor);
+					return r;
+				}
+			}
+		}
+		CLEARDBT(key);	
+		CLEARDBT(data);	
+	}
+	if( dbret != DB_NOTFOUND ) {
+		db->database->err(db->database, dbret, "files.db:");
+		return RET_DBERR(dbret);
+	}
+	if( (dbret = cursor->c_close(cursor)) != 0 ) {
+		db->database->err(db->database, dbret, "files.db:");
+		return RET_DBERR(dbret);
+	}
+	return result;
+}
