@@ -17,10 +17,13 @@
 	IGN(unknownfield) \
 	IGN(wrongdistribution) \
 	IGN(missingfield) \
+	IGN(doublefield) \
 	IGN(brokenold) \
 	IGN(brokenversioncmp) \
+	IGN(extension) \
 	IGN(unusedarch) \
-	IGN(surprisingarch)
+	IGN(surprisingarch) \
+	IGN(missingfile)
 
 
 enum ignore { 
