@@ -47,4 +47,11 @@ retvalue propercodename(const char *codename);
 retvalue properarchitectures(const struct strlist *architectures);
 retvalue propercomponents(const struct strlist *components);
 
+/* create two stringlists out of one,
+ * from[i]=into[i]=list[i] if list[i] contains no '>'
+ * from[i] '>' into[i] = list[i] when list[i] contains a '>'
+ */
+retvalue splitlist(/*@out@*/struct strlist *from,
+				/*@out@*/struct strlist *into,
+				const struct strlist *list);
 #endif
