@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
  *  Copyright (C) 2003,2004,2005 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -146,8 +145,7 @@ retvalue release_init(const char *dbdir, const char *distdir, const char *codena
 		(void)n->cachedb->close(n->cachedb,0);
 		return RET_ERROR_OOM;
 	}
-	dbret = n->cachedb->open(n->cachedb,filename,codename,
-			DB_HASH, DB_CREATE, 0664);
+	dbret = DB_OPEN(n->cachedb,filename,codename, DB_HASH, DB_CREATE);
 	if( dbret < 0 ) {
 		n->cachedb->err(n->cachedb, dbret, "%s(%s)",
 				filename,codename);
@@ -282,14 +280,11 @@ static retvalue release_usecached(struct release *release,
 	if( filename == NULL )
 		return RET_ERROR_OOM;
 
-	if( (compressions & IC_FLAG(ic_uncompressed)) != 0 ) {
-		r = getcachevalue(release,relfilename,&md5sum);
-		if( !RET_IS_OK(r) ) {
-			free(filename);
-			return r;
-		}
-	} else
-		md5sum = NULL;
+	r = getcachevalue(release,relfilename,&md5sum);
+	if( !RET_IS_OK(r) ) {
+		free(filename);
+		return r;
+	}
 	if( (compressions & IC_FLAG(ic_gzip)) != 0 ) {
 		gzrelfilename = calc_addsuffix(relfilename,"gz");
 		if( gzrelfilename == NULL ) {
@@ -1053,6 +1048,7 @@ retvalue release_directorydescription(struct release *release, const struct dist
 	release_writeheader("Origin",distribution->origin);
 	release_writeheader("Label",distribution->label);
 	release_writeheader("Architecture",target->architecture);
+	release_writeheader("NotAutomatic",distribution->notautomatic);
 	release_writeheader("Description",distribution->description);
 	r = release_finishfile(release, f);
 	return r;
@@ -1240,6 +1236,12 @@ retvalue release_write(/*@only@*/struct release *release, struct distribution *d
 		e = fputs(distribution->description,f);
 		checkwritten;
 	}
+	if( distribution->notautomatic != NULL ) {
+		e = fputs("\nNotAutomatic: ",f);
+		checkwritten;
+		e = fputs(distribution->notautomatic,f);
+		checkwritten;
+	}
 
 	e = fputs("\nMD5Sum:\n",f);
 	checkwritten;
@@ -1354,7 +1356,7 @@ retvalue release_write(/*@only@*/struct release *release, struct distribution *d
 		RET_UPDATE(result,r);
 	}
 
-	/* now update the cache database, do we find those the next time */
+	/* now update the cache database, so we find those the next time */
 
 	r = storechecksums(release);
 	RET_UPDATE(result,r);
