@@ -6,6 +6,10 @@
 #warning "What's hapening here?"
 #endif
 
+/* does not need to be called if allowpassphrase if FALSE,
+ * argument will only take effect if called the first time */
+retvalue signature_init(bool_t allowpassphrase);
+
 retvalue signature_check(const char *options, const char *releasegpg, const char *release);
 retvalue signature_sign(const char *options, const char *filename, const char *signeturename);
 
