@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2003,2004,2005 Bernhard R. Link
+ *  Copyright (C) 2003,2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -349,6 +348,9 @@ retvalue sources_doreoverride(const struct alloverrides *ao,const char *packagen
 	struct fieldtoadd *fields;
 	char *newchunk;
 
+	if( interupted() )
+		return RET_ERROR_INTERUPTED;
+
 	o = override_search(ao->dsc, packagename);
 	if( o == NULL )
 		return RET_NOTHING;
@@ -373,6 +375,9 @@ retvalue sources_retrack(struct target *t,const char *sourcename,const char *chu
 	//TODO: elliminate duplicate code!
 	assert(sourcename!=NULL);
 
+	if( interupted() )
+		return RET_ERROR_INTERUPTED;
+
 	r = chunk_getvalue(chunk,"Version",&sourceversion);
 	if( r == RET_NOTHING ) {
 		fprintf(stderr,"Did not find Version in chunk:'%s'\n",chunk);
@@ -392,36 +397,19 @@ retvalue sources_retrack(struct target *t,const char *sourcename,const char *chu
 		return r;
 	}
 
-	r = tracking_get(tracks,sourcename,sourceversion,&pkg);
+	r = tracking_getornew(tracks,sourcename,sourceversion,&pkg);
+	free(sourceversion);
 	if( RET_WAS_ERROR(r) ) {
-		free(sourceversion);
 		strlist_done(&filekeys);
 		return r;
 	}
-	if( r == RET_NOTHING ) {
-		r = tracking_new(tracks,sourcename,sourceversion,&pkg);
-		free(sourceversion);
-		if( RET_WAS_ERROR(r) ) {
-			strlist_done(&filekeys);
-			return r;
-		}
-		r = trackedpackage_addfilekeys(tracks,pkg,ft_SOURCE,&filekeys,TRUE,refs);
-		strlist_done(&filekeys);
-		if( RET_WAS_ERROR(r) )
-			return r;
-		r = tracking_put(tracks,pkg);
+
+	r = trackedpackage_addfilekeys(tracks,pkg,ft_SOURCE,&filekeys,TRUE,refs);
+	if( RET_WAS_ERROR(r) ) {
 		trackedpackage_free(pkg);
 		return r;
 	}
-	free(sourceversion);
-
-	r = trackedpackage_addfilekeys(tracks,pkg,ft_SOURCE,&filekeys,TRUE,refs);
-	strlist_done(&filekeys);
-	if( RET_WAS_ERROR(r) )
-		return r;
-	r = tracking_replace(tracks,pkg);
-	trackedpackage_free(pkg);
-	return r;
+	return tracking_save(tracks, pkg);
 }
 
 retvalue sources_getsourceandversion(UNUSED(struct target *t),const char *chunk,const char *packagename,char **source,char **version) {
