@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
  *  Copyright (C) 2003,2005 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -38,6 +37,20 @@ bool_t strlist_in(const struct strlist *strlist,const char *element) {
 	}
 	return FALSE;
 }
+int strlist_ofs(const struct strlist *strlist,const char *element) {
+	int c;
+	char **t;
+
+	assert(strlist != NULL);
+
+	c = strlist->count; 
+	t = strlist->values;
+	while( c-- != 0 ) {
+		if( strcmp(*(t++),element) == 0 )
+			return (t-strlist->values)-1;
+	}
+	return -1;
+}
 
 bool_t strlist_subset(const struct strlist *strlist,const struct strlist *subset,const char **missing) {
 	int c;
