@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2003,2004,2005 Bernhard R. Link
+ *  Copyright (C) 2003,2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -44,6 +43,7 @@
 #include "terms.h"
 #include "filterlist.h"
 #include "readrelease.h"
+#include "donefile.h"
 
 // TODO: what about other signatures? Is hard-coding ".gpg" sensible?
 
@@ -142,6 +142,10 @@ struct update_index {
 	struct update_origin *origin;
 	char *filename;
 	char *upstream;
+	/* != NULL if filename was changed by ListHook, then the original */
+	char *original_filename;
+	/* index into origin's checkfile, -1 = none */
+	int checksum_ofs;
 	/* there was something missed here */
 	bool_t failed;
 };
@@ -155,6 +159,8 @@ struct update_target {
 	bool_t ignoredelete;
 	/* don't do anything because of --skipold */
 	bool_t nothingnew;
+	/* if true do not generate donefiles */
+	bool_t incomplete;
 };
 
 struct update_distribution {
@@ -221,6 +227,7 @@ static void updates_freetargets(/*@only@*/struct update_target *t) {
 			ui = ut->indices;
 			ut->indices = ui->next;
 			free(ui->filename);
+			free(ui->original_filename);
 			free(ui->upstream);
 			free(ui);
 		}
@@ -252,71 +259,11 @@ static inline retvalue newupdatetarget(struct update_target **ts,/*@dependent@*/
 	ut->upgradelist = NULL;
 	ut->ignoredelete = FALSE;
 	ut->nothingnew = FALSE;
+	ut->incomplete = FALSE;
 	*ts = ut;
 	return RET_OK;
 }
 
-static retvalue splitlist(struct strlist *from,
-				struct strlist *into,
-				const struct strlist *list) {
-	retvalue r;
-	int i;
-
-	r = strlist_init(from);
-	if( RET_WAS_ERROR(r) ) {
-		return r;
-	}
-	r = strlist_init(into);
-	if( RET_WAS_ERROR(r) ) {
-		strlist_done(from);
-		return r;
-	}
-
-	/* * Iterator over components to update * */
-	r = RET_NOTHING;
-	for( i = 0 ; i < list->count ; i++ ) {
-		const char *item,*seperator;
-		char *origin,*destination;
-
-		item = list->values[i];
-		// TODO: isn't this broken for the !*(dest+1) case ?
-		if( (seperator = strchr(item,'>')) == NULL ) {
-			destination = strdup(item);
-			origin = strdup(item);
-		} else if( seperator == item ) {
-			destination = strdup(seperator+1);
-			origin = strdup(seperator+1);
-		} else if( *(seperator+1) == '\0' ) {
-			destination = strndup(item,seperator-item);
-			origin = strndup(item,seperator-item);
-		} else {
-			origin = strndup(item,seperator-item);
-			destination = strdup(seperator+1);
-		}
-		if( origin == NULL || destination == NULL ) {
-			free(origin);free(destination);
-			strlist_done(from);
-			strlist_done(into);
-			return RET_ERROR_OOM;
-		}
-		r = strlist_add(from,origin);
-		if( RET_WAS_ERROR(r) ) {
-			free(destination);
-			strlist_done(from);
-			strlist_done(into);
-			return r;
-		}
-		r = strlist_add(into,destination);
-		if( RET_WAS_ERROR(r) ) {
-			strlist_done(from);
-			strlist_done(into);
-			return r;
-		}
-		r = RET_OK;
-	}
-	return r;
-}
-
 inline static retvalue parse_pattern(const char *confdir,const char *chunk, struct update_pattern **pattern) {
 	struct update_pattern *update;
 	struct strlist componentslist,architectureslist;
@@ -608,7 +555,7 @@ retvalue updates_getpatterns(const char *confdir,struct update_pattern **pattern
 		return RET_ERROR_OOM;
 	data.patterns = &update;
 	data.confdir = confdir;
-	r = chunk_foreach(updatesfile,parsechunk,&data,0,FALSE);
+	r = chunk_foreach(updatesfile,parsechunk,&data,FALSE);
 	free(updatesfile);
 	if( RET_IS_OK(r) )
 		*patterns = update;
@@ -692,6 +639,7 @@ static inline retvalue newindex(struct update_index **indices,
 		free(index);
 		return RET_ERROR_OOM;
 	}
+	index->original_filename = NULL;
 	index->upstream = (*target->getupstreamindex)(target,
 		origin->suite_from,component_from,architecture_from);
 	if( index->upstream == NULL ) {
@@ -700,6 +648,7 @@ static inline retvalue newindex(struct update_index **indices,
 		return RET_ERROR_OOM;
 	}
 	index->origin = origin;
+	index->checksum_ofs = -1;
 	index->next = *indices;
 	index->failed = FALSE;
 	*indices = index;
@@ -913,10 +862,15 @@ static bool_t foundinindices(struct update_target *targets,size_t nameoffset,con
 
 	for( t = targets; t != NULL ; t = t->next ) {
 		for( i = t->indices ; i != NULL ; i=i->next ) {
+			size_t l;
+
 			if( i->filename == NULL )
 				continue;
-			assert(strlen(i->filename) > nameoffset);
-			if( strcmp(name,i->filename+nameoffset) == 0 )
+			l = strlen(i->filename);
+			assert(l > nameoffset);
+			if( strncmp(name,i->filename+nameoffset,l-nameoffset) == 0 &&
+				(name[l-nameoffset] == '\0' ||
+				 strcmp(name+(l-nameoffset), ".done") == 0) )
 				return TRUE;
 		}
 	}
@@ -1023,7 +977,7 @@ static inline retvalue startuporigin(struct aptmethodrun *run,struct update_orig
 	return RET_OK;
 }
 
-static retvalue updates_startup(struct aptmethodrun *run,struct update_distribution *distributions, int force) {
+static retvalue updates_startup(struct aptmethodrun *run,struct update_distribution *distributions) {
 	retvalue result,r;
 	struct update_origin *origin;
 	struct update_distribution *d;
@@ -1037,15 +991,14 @@ static retvalue updates_startup(struct aptmethodrun *run,struct update_distribut
 				fprintf(stderr,"Warning: Override-Files of '%s' ignored as not yet supported while updating!\n",d->distribution->codename);
 		}
 		if( d->distribution->tracking != dt_NONE ) {
-			fprintf(stderr,"Error(%s): Coping with trackingdata while updating not yet implemented!\n",d->distribution->codename);
-			return RET_ERROR;
+			fprintf(stderr,"WARNING: Updating does not update trackingdata. Trackingdata of %s will be outdated!\n",d->distribution->codename);
 		}
 		for( origin=d->origins; origin != NULL ; origin=origin->next ) {
 			if( origin->pattern == NULL)
 				continue;
 			r = startuporigin(run,origin);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				return r;
 		}
 	}
@@ -1110,19 +1063,26 @@ static inline retvalue readchecksums(struct update_origin *origin) {
 	return r;
 }
 
-/* returns RET_NOTHING when nothing new will be retrieved */
-static inline retvalue queueindex(struct update_index *index,int force) {
+/* returns RET_NOTHING when nothing new will be there */
+static inline retvalue queueindex(struct update_index *index) {
 	const struct update_origin *origin = index->origin;
 	int i;
 	size_t l;
 
 	assert( index != NULL && index->origin != NULL );
+	index->checksum_ofs = -1;
 	if( origin->download == NULL || origin->failed ) {
 		return RET_NOTHING;
 	}
-	if( origin->releasefile == NULL )
-		return aptmethod_queueindexfile(origin->download,
-			index->upstream,index->filename);
+	if( origin->releasefile == NULL ) {
+		retvalue r;
+		/* TODO: save the content to make sure it is the old instead */
+		donefile_delete(index->filename);
+		r = aptmethod_queueindexfile(origin->download,
+				index->upstream,index->filename);
+		assert( r != RET_NOTHING );
+		return r;
+	}
 
 	//TODO: this is a very crude hack, think of something better...
 	l = 7 + strlen(origin->suite_from); // == strlen("dists/%s/")
@@ -1134,31 +1094,31 @@ static inline retvalue queueindex(struct update_index *index,int force) {
 			retvalue r;
 			const char *md5sum = origin->checksums.values[i+1];
 
+			index->checksum_ofs = i+1;
+
 			r = md5sum_ensure(index->filename,md5sum,FALSE);
 			if( r == RET_NOTHING ) {
+				donefile_delete(index->filename);
 				r = aptmethod_queuefile(origin->download,
 					index->upstream,index->filename,
 					md5sum,NULL,NULL);
 			} else if( RET_IS_OK(r) ) {
-				/* file is already there, nothing has
-				 * to be done */
-				r = RET_NOTHING;
+				/* file is already there, but it might still need
+				 * processing as last time it was not due to some
+				 * error of some other file */
+				r = donefile_isold(index->filename, md5sum);
 			}
 			return r;
 		}
 		
 	}
 	fprintf(stderr,"Could not find '%s' within the Releasefile of '%s':\n'%s'\n",index->upstream,origin->pattern->name,origin->releasefile);
-	if( force > 2 ) {
-		aptmethod_queueindexfile(origin->download,
-				index->upstream,index->filename);
-	}
 	return RET_ERROR_WRONG_MD5;
 }
 
 
 
-static retvalue updates_queuemetalists(struct update_distribution *distributions, int force) {
+static retvalue updates_queuemetalists(struct update_distribution *distributions) {
 	retvalue result,r;
 	struct update_origin *origin;
 	struct update_distribution *d;
@@ -1172,7 +1132,7 @@ static retvalue updates_queuemetalists(struct update_distribution *distributions
 				continue;
 			r = queuemetalists(origin);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				return r;
 		}
 
@@ -1181,7 +1141,7 @@ static retvalue updates_queuemetalists(struct update_distribution *distributions
 	return result;
 }
 
-static retvalue updates_queuelists(struct update_distribution *distributions,bool_t skipold,bool_t *anythingtodo,int force) {
+static retvalue updates_queuelists(struct update_distribution *distributions,bool_t skipold,bool_t *anythingtodo) {
 	retvalue result,r;
 	struct update_origin *origin;
 	struct update_target *target;
@@ -1196,20 +1156,17 @@ static retvalue updates_queuelists(struct update_distribution *distributions,boo
 				continue;
 			r = readchecksums(origin);
 			RET_UPDATE(result,r);
-			//TODO: more force needed?
-			if( RET_WAS_ERROR(r) && force <= 1 ) {
-				if( force == 1 ) {
-					fprintf(stderr,"To ignore errors checking Release files specify --force two times.\n");
-				}
+			if( RET_WAS_ERROR(r) ) {
 				return r;
 			}
 		}
 		for( target=d->targets; target!=NULL ; target=target->next ) {
 			target->nothingnew = skipold;
+			target->incomplete = FALSE;
 			for( index=target->indices ; index!=NULL ; index=index->next ) {
 				if( index->origin == NULL )
 					continue;
-				r = queueindex(index,force);
+				r = queueindex(index);
 				if( RET_IS_OK(r) ) {
 					target->nothingnew = FALSE;
 					*anythingtodo = TRUE;
@@ -1217,7 +1174,7 @@ static retvalue updates_queuelists(struct update_distribution *distributions,boo
 				if( RET_WAS_ERROR(r) )
 					index->failed = TRUE;
 				RET_UPDATE(result,r);
-				if( RET_WAS_ERROR(r) && force <= 0 )
+				if( RET_WAS_ERROR(r) )
 					return r;
 			}
 		}
@@ -1254,8 +1211,9 @@ static retvalue calllisthook(const char *listhook,struct update_index *index) {
 		exit(255);
 	}
 	if( verbose > 5 )
-		fprintf(stderr,"Called %s '%s# '%s'\n",listhook,index->filename,newfilename);
-	free(index->filename);
+		fprintf(stderr,"Called %s '%s' '%s'\n",listhook,index->filename,newfilename);
+	assert(index->original_filename == NULL);
+	index->original_filename = index->filename;
 	index->filename=newfilename;
 	do {
 		c = waitpid(f,&status,WUNTRACED);
@@ -1280,7 +1238,7 @@ static retvalue calllisthook(const char *listhook,struct update_index *index) {
 	}
 }
 
-static retvalue updates_calllisthooks(struct update_distribution *distributions,int force) {
+static retvalue updates_calllisthooks(struct update_distribution *distributions) {
 	retvalue result,r;
 	struct update_target *target;
 	struct update_index *index;
@@ -1302,7 +1260,7 @@ static retvalue updates_calllisthooks(struct update_distribution *distributions,
 					continue;
 				}
 				r = calllisthook(index->origin->pattern->listhook,index);
-				if( RET_WAS_ERROR(r) && force <= 0 ) {
+				if( RET_WAS_ERROR(r) ) {
 					index->failed = TRUE;
 					return r;
 				}
@@ -1343,7 +1301,7 @@ static upgrade_decision ud_decide_by_pattern(void *privdata, const char *package
 	return UD_UPGRADE;
 }
 
-static inline retvalue searchformissing(const char *dbdir,struct update_target *u,int force) {
+static inline retvalue searchformissing(const char *dbdir,struct update_target *u) {
 	struct update_index *index;
 	retvalue result,r;
 
@@ -1367,8 +1325,10 @@ static inline retvalue searchformissing(const char *dbdir,struct update_target *
 			if( verbose > 4 )
 				fprintf(stderr,"  marking everything to be deleted\n");
 			r = upgradelist_deleteall(u->upgradelist);
+			if( RET_WAS_ERROR(r) )
+				u->incomplete = TRUE;
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				return result;
 			u->ignoredelete = FALSE;
 			continue;
@@ -1377,6 +1337,7 @@ static inline retvalue searchformissing(const char *dbdir,struct update_target *
 		if( index->failed || index->origin->failed ) {
 			if( verbose >= 1 )
 				fprintf(stderr,"  missing '%s'\n",index->filename);
+			u->incomplete = TRUE;
 			u->ignoredelete = TRUE;
 			continue;
 		}
@@ -1387,33 +1348,36 @@ static inline retvalue searchformissing(const char *dbdir,struct update_target *
 		r = upgradelist_update(u->upgradelist,
 				index->origin->download,index->filename,
 				ud_decide_by_pattern,
-				(void*)index->origin->pattern,
-				force);
-		if( RET_WAS_ERROR(r) )
+				(void*)index->origin->pattern);
+		if( RET_WAS_ERROR(r) ) {
+			u->incomplete = TRUE;
 			u->ignoredelete = TRUE;
+		}
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			return result;
 	}
 	
 	return result;
 }
 
-static retvalue updates_readindices(const char *dbdir,struct update_distribution *d,int force) {
+static retvalue updates_readindices(const char *dbdir,struct update_distribution *d) {
 	retvalue result,r;
 	struct update_target *u;
 
 	result = RET_NOTHING;
 	for( u=d->targets ; u != NULL ; u=u->next ) {
-		r = searchformissing(dbdir,u,force);
+		r = searchformissing(dbdir,u);
+		if( RET_WAS_ERROR(r) )
+			u->incomplete = TRUE;
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 	}
 	return result;
 }
 
-static retvalue updates_enqueue(struct downloadcache *cache,filesdb filesdb,struct update_distribution *distribution,int force) {
+static retvalue updates_enqueue(struct downloadcache *cache,filesdb filesdb,struct update_distribution *distribution) {
 	retvalue result,r;
 	struct update_target *u;
 
@@ -1421,16 +1385,18 @@ static retvalue updates_enqueue(struct downloadcache *cache,filesdb filesdb,stru
 	for( u=distribution->targets ; u != NULL ; u=u->next ) {
 		if( u->nothingnew )
 			continue;
-		r = upgradelist_enqueue(u->upgradelist,cache,filesdb,force);
+		r = upgradelist_enqueue(u->upgradelist,cache,filesdb);
+		if( RET_WAS_ERROR(r) )
+			u->incomplete = TRUE;
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 	}
 	return result;
 }
 
 
-static retvalue updates_install(const char *dbdir,filesdb filesdb,references refs,struct update_distribution *distribution,int force,struct strlist *dereferencedfilekeys) {
+static retvalue updates_install(const char *dbdir,filesdb filesdb,references refs,struct update_distribution *distribution,struct strlist *dereferencedfilekeys) {
 	retvalue result,r;
 	struct update_target *u;
 
@@ -1438,57 +1404,92 @@ static retvalue updates_install(const char *dbdir,filesdb filesdb,references ref
 	for( u=distribution->targets ; u != NULL ; u=u->next ) {
 		if( u->nothingnew )
 			continue;
-		r = upgradelist_install(u->upgradelist,dbdir,filesdb,refs,force,u->ignoredelete,dereferencedfilekeys);
+		r = upgradelist_install(u->upgradelist,dbdir,filesdb,refs,u->ignoredelete,dereferencedfilekeys);
+		RET_UPDATE(distribution->distribution->status, r);
+		if( RET_WAS_ERROR(r) )
+			u->incomplete = TRUE;
 		RET_UPDATE(result,r);
 		upgradelist_free(u->upgradelist);
 		u->upgradelist = NULL;
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 	}
 	return result;
 }
 
+static void markdone(struct update_target *target) {
+	struct update_index *index;
+
+	if( target->incomplete )
+		return;
+
+	for( index=target->indices ; index != NULL ; index=index->next ) {
+		const struct update_origin *origin = index->origin;
+		const char *md5sum;
+
+		if( origin == NULL )
+			/* No need to mark a delete rule as done */
+			continue;
+		if( origin->releasefile == NULL )
+			/* TODO: once they can be detected as
+			 * old, also generate donefile here */
+			continue;
+		if( index->checksum_ofs < 0 )
+			continue;
+		assert( index->checksum_ofs < origin->checksums.count );
+		md5sum = origin->checksums.values[index->checksum_ofs];
+		if( index->original_filename == NULL )
+			donefile_create(index->filename, md5sum);
+		else
+			donefile_create(index->original_filename, md5sum);
+	}
+}
+
 static void updates_dump(struct update_distribution *distribution) {
 	struct update_target *u;
 
 	for( u=distribution->targets ; u != NULL ; u=u->next ) {
+		if( u->nothingnew )
+			continue;
 		printf("Updates needed for '%s':\n",u->target->identifier);
 		upgradelist_dump(u->upgradelist);
+		upgradelist_free(u->upgradelist);
+		u->upgradelist = NULL;
 	}
 }
 
-static retvalue updates_downloadlists(const char *methoddir,struct aptmethodrun *run,struct update_distribution *distributions, bool_t skipold, bool_t *anythingtodo, int force) {
+static retvalue updates_downloadlists(const char *methoddir,struct aptmethodrun *run,struct update_distribution *distributions, bool_t skipold, bool_t *anythingtodo) {
 	retvalue r,result;
 
 	/* first get all "Release" and "Release.gpg" files */
-	result = updates_queuemetalists(distributions,force);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	result = updates_queuemetalists(distributions);
+	if( RET_WAS_ERROR(result) ) {
 		return result;
 	}
 
 	r = aptmethod_download(run,methoddir,NULL);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(r) && force <= 0 ) {
+	if( RET_WAS_ERROR(r) ) {
 		return result;
 	}
 
 	/* Then get all index files (with perhaps md5sums from the above) */
-	r = updates_queuelists(distributions,skipold,anythingtodo,force);
+	r = updates_queuelists(distributions,skipold,anythingtodo);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
 		return result;
 	}
 
 	r = aptmethod_download(run,methoddir,NULL);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(r) && force <= 0 ) {
+	if( RET_WAS_ERROR(r) ) {
 		return result;
 	}
 
 	return result;
 }
 
-retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *distributions,int force,bool_t nolistsdownload,bool_t skipold,struct strlist *dereferencedfilekeys) {
+retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *distributions,bool_t nolistsdownload,bool_t skipold,struct strlist *dereferencedfilekeys) {
 	retvalue result,r;
 	struct update_distribution *d;
 	struct aptmethodrun *run;
@@ -1506,8 +1507,8 @@ retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,
 	}
 
 	/* preperations */
-	result = updates_startup(run,distributions,force);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	result = updates_startup(run,distributions);
+	if( RET_WAS_ERROR(result) ) {
 		aptmethod_shutdown(run);
 		return result;
 	}
@@ -1517,23 +1518,31 @@ retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,
 	} else {
 		bool_t anythingtodo = !skipold;
 
-		r = updates_downloadlists(methoddir,run,distributions,skipold,&anythingtodo,force);
+		r = updates_downloadlists(methoddir,run,distributions,skipold,&anythingtodo);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(result) && force <= 0 ) {
+		if( RET_WAS_ERROR(result) ) {
 			aptmethod_shutdown(run);
 			return result;
 		}
+		/* TODO: 
+		 * add a check if some of the upstreams without Release files
+		 * are unchanged and if this changes anything? */
 		if( !anythingtodo ) {
 			fprintf(stderr,"Nothing to do found. (Use --noskipold to force processing)\n");
-			return result;
+			aptmethod_shutdown(run);
+			if( RET_IS_OK(result) )
+				return RET_NOTHING;
+			else
+				return result;
 		}
 	}
 	/* Call ListHooks (if given) on the downloaded index files.
 	 * (This is done even when nolistsdownload is given, as otherwise
 	 *  the filename to look in is not changed) */
-	r = updates_calllisthooks(distributions,force);
+	r = updates_calllisthooks(distributions);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
+		aptmethod_shutdown(run);
 		return result;
 	}
 
@@ -1548,16 +1557,25 @@ retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,
 	}
 	
 	for( d=distributions ; d != NULL ; d=d->next) {
-		r = updates_readindices(dbdir,d,force);
+		r = updates_readindices(dbdir,d);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
-		r = updates_enqueue(cache,filesdb,d,force);
+		r = updates_enqueue(cache,filesdb,d);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 	}
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
+		for( d=distributions ; d != NULL ; d=d->next) {
+			struct update_target *u;
+			for( u=d->targets ; u != NULL ; u=u->next ) {
+				upgradelist_free(u->upgradelist);
+				u->upgradelist = NULL;
+			}
+		}
+		r = downloadcache_free(cache);
+		RET_UPDATE(result,r);
 		aptmethod_shutdown(run);
 		return result;
 	}
@@ -1574,23 +1592,38 @@ retvalue updates_update(const char *dbdir,const char *methoddir,filesdb filesdb,
 	r = aptmethod_shutdown(run);
 	RET_UPDATE(result,r);
 
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
+		for( d=distributions ; d != NULL ; d=d->next) {
+			struct update_target *u;
+			for( u=d->targets ; u != NULL ; u=u->next ) {
+				upgradelist_free(u->upgradelist);
+				u->upgradelist = NULL;
+			}
+		}
 		return result;
 	}
 	if( verbose >= 0 )
 		fprintf(stderr,"Installing (and possibly deleting) packages...\n");
 
 	for( d=distributions ; d != NULL ; d=d->next) {
-		r = updates_install(dbdir,filesdb,refs,d,force,dereferencedfilekeys);
+		r = updates_install(dbdir,filesdb,refs,d,dereferencedfilekeys);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 	}
 
+	for( d=distributions ; d != NULL ; d=d->next) {
+		struct update_target *u;
+
+		for( u=d->targets ; u != NULL ; u=u->next ) {
+			markdone(u);
+		}
+	}
+
 	return result;
 }
 
-retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct update_distribution *distributions,int force,bool_t nolistsdownload,bool_t skipold) {
+retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct update_distribution *distributions,bool_t nolistsdownload,bool_t skipold) {
 	struct update_distribution *d;
 	retvalue result,r;
 	struct aptmethodrun *run;
@@ -1606,8 +1639,8 @@ retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct upda
 		skipold = FALSE;
 	}
 
-	result = updates_startup(run,distributions,force);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	result = updates_startup(run,distributions);
+	if( RET_WAS_ERROR(result) ) {
 		aptmethod_shutdown(run);
 		return result;
 	}
@@ -1616,9 +1649,9 @@ retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct upda
 			fprintf(stderr,"Warning: As --nolistsdownload is given, index files are NOT checked.\n");
 	} else {
 		bool_t anythingtodo = !skipold;
-		r = updates_downloadlists(methoddir,run,distributions,skipold,&anythingtodo,force);
+		r = updates_downloadlists(methoddir,run,distributions,skipold,&anythingtodo);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(result) && force <= 0 ) {
+		if( RET_WAS_ERROR(result) ) {
 			aptmethod_shutdown(run);
 			return result;
 		}
@@ -1630,16 +1663,16 @@ retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct upda
 	/* Call ListHooks (if given) on the downloaded index files.
 	 * (This is done even when nolistsdownload is given, as otherwise
 	 *  the filename to look in is not changed) */
-	r = updates_calllisthooks(distributions,force);
+	r = updates_calllisthooks(distributions);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
 		return result;
 	}
 	if( verbose > 0 )
 		fprintf(stderr,"Shutting down aptmethods...\n");
 	r = aptmethod_shutdown(run);
 	RET_UPDATE(result,r);
-	if( RET_WAS_ERROR(result) && force <= 0 ) {
+	if( RET_WAS_ERROR(result) ) {
 		return result;
 	}
 
@@ -1648,9 +1681,9 @@ retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct upda
 		fprintf(stderr,"Calculating packages to get...\n");
 	
 	for( d=distributions ; d != NULL ; d=d->next) {
-		r = updates_readindices(dbdir,d,force);
+		r = updates_readindices(dbdir,d);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 )
+		if( RET_WAS_ERROR(r) )
 			break;
 		updates_dump(d);
 	}
@@ -1658,7 +1691,103 @@ retvalue updates_checkupdate(const char *dbdir,const char *methoddir,struct upda
 	return result;
 }
 
-static retvalue singledistributionupdate(const char *dbdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *d,int force,bool_t nolistsdownload,bool_t skipold, struct strlist *dereferencedfilekeys) {
+retvalue updates_predelete(const char *dbdir,const char *methoddir,references refs,struct update_distribution *distributions,bool_t nolistsdownload,bool_t skipold,struct strlist *dereferencedfilekeys) {
+	retvalue result,r;
+	struct update_distribution *d;
+	struct aptmethodrun *run;
+
+	r = aptmethod_initialize_run(&run);
+	if( RET_WAS_ERROR(r) )
+		return r;
+
+	if( nolistsdownload ) {
+		if( skipold && verbose >= 0 ) {
+			fprintf(stderr,"Ignoring --skipold because of --nolistsdownload\n");
+		}
+		skipold = FALSE;
+	}
+
+	/* preperations */
+	result = updates_startup(run,distributions);
+	if( RET_WAS_ERROR(result) ) {
+		aptmethod_shutdown(run);
+		return result;
+	}
+	if( nolistsdownload ) {
+		if( verbose >= 0 )
+			fprintf(stderr,"Warning: As --nolistsdownload is given, index files are NOT checked.\n");
+	} else {
+		bool_t anythingtodo = !skipold;
+
+		r = updates_downloadlists(methoddir,run,distributions,skipold,&anythingtodo);
+		RET_UPDATE(result,r);
+		if( RET_WAS_ERROR(result) ) {
+			aptmethod_shutdown(run);
+			return result;
+		}
+		/* TODO: 
+		 * add a check if some of the upstreams without Release files
+		 * are unchanged and if this changes anything? */
+		if( !anythingtodo ) {
+			fprintf(stderr,"Nothing to do found. (Use --noskipold to force processing)\n");
+			aptmethod_shutdown(run);
+			if( RET_IS_OK(result) )
+				return RET_NOTHING;
+			else
+				return result;
+		}
+	}
+	/* Call ListHooks (if given) on the downloaded index files.
+	 * (This is done even when nolistsdownload is given, as otherwise
+	 *  the filename to look in is not changed) */
+	r = updates_calllisthooks(distributions);
+	RET_UPDATE(result,r);
+	if( RET_WAS_ERROR(result) ) {
+		aptmethod_shutdown(run);
+		return result;
+	}
+
+	if( verbose > 0 )
+		fprintf(stderr,"Shutting down aptmethods...\n");
+
+	r = aptmethod_shutdown(run);
+	RET_UPDATE(result,r);
+	if( RET_WAS_ERROR(result) ) {
+		return result;
+	}
+	
+	if( verbose >= 0 )
+		fprintf(stderr,"Removing obsolete or to be replaced packages...\n");
+	for( d=distributions ; d != NULL ; d=d->next) {
+		struct update_target *u;
+
+		for( u=d->targets ; u != NULL ; u=u->next ) {
+			r = searchformissing(dbdir,u);
+			RET_UPDATE(result,r);
+			if( RET_WAS_ERROR(r) ) {
+				u->incomplete = TRUE;
+				continue;
+			}
+			if( u->nothingnew || u->ignoredelete ) {
+				upgradelist_free(u->upgradelist);
+				u->upgradelist = NULL;
+				continue;
+			}
+			r = upgradelist_predelete(u->upgradelist,dbdir,refs,dereferencedfilekeys);
+			RET_UPDATE(d->distribution->status, r);
+			if( RET_WAS_ERROR(r) )
+				u->incomplete = TRUE;
+			RET_UPDATE(result,r);
+			upgradelist_free(u->upgradelist);
+			u->upgradelist = NULL;
+			if( RET_WAS_ERROR(r) )
+				return r;
+		}
+	}
+	return result;
+}
+
+static retvalue singledistributionupdate(const char *dbdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *d,bool_t nolistsdownload,bool_t skipold, struct strlist *dereferencedfilekeys) {
 	struct aptmethodrun *run;
 	struct downloadcache *cache;
 	struct update_origin *origin;
@@ -1677,7 +1806,7 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 			continue;
 		r = startuporigin(run,origin);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			aptmethod_shutdown(run);
 			return r;
 		}
@@ -1691,16 +1820,14 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 			r = queuemetalists(origin);
 			RET_UPDATE(result,r);
 			if( RET_WAS_ERROR(r) ) {
-				if( force <= 0 ) {
-					aptmethod_shutdown(run);
-					return r;
-				}
+				aptmethod_shutdown(run);
 				origin->failed = TRUE;
+				return r;
 			}
 		}
 		r = aptmethod_download(run,methoddir,NULL);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			aptmethod_shutdown(run);
 			return r;
 		}
@@ -1709,10 +1836,7 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 				continue;
 			r = readchecksums(origin);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 1 ) {
-				if( force == 1 ) {
-					fprintf(stderr,"To ignore errors checking Release files specify --force two times.\n");
-				}
+			if( RET_WAS_ERROR(r) ) {
 				aptmethod_shutdown(run);
 				return r;
 			}
@@ -1722,18 +1846,17 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 	for( target=d->targets; target!=NULL ; target=target->next ) {
 		struct update_index *index;
 		target->nothingnew = skipold;
+		target->incomplete = FALSE;
 		if( !nolistsdownload ) {
 			for( index=target->indices ; index!=NULL ; index=index->next ) {
 				if( index->origin == NULL || index->origin->failed )
 					continue;
-				r = queueindex(index,force);
+				r = queueindex(index);
 				if( RET_WAS_ERROR(r) ) {
 					RET_UPDATE(result,r);
-					if( force <= 0 ) {
-						aptmethod_shutdown(run);
-						return r;
-					}
+					aptmethod_shutdown(run);
 					index->failed = TRUE;
+					return r;
 				} else if( RET_IS_OK(r) ) {
 					target->nothingnew = FALSE;
 				}
@@ -1742,7 +1865,7 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 				continue;
 			r = aptmethod_download(run,methoddir,NULL);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 ) {
+			if( RET_WAS_ERROR(r) ) {
 				aptmethod_shutdown(run);
 				return r;
 			}
@@ -1761,10 +1884,8 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 			if( RET_WAS_ERROR(r) ) {
 				index->failed = TRUE;
 				RET_UPDATE(result,r);
-				if( force <= 0) {
-					aptmethod_shutdown(run);
-					return r;
-				}
+				aptmethod_shutdown(run);
+				return r;
 			}
 
 		}
@@ -1775,26 +1896,30 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 		RET_UPDATE(result,r);	
 		if( !RET_IS_OK(r) ) {
 			(void)downloadcache_free(cache);
-			if( RET_WAS_ERROR(r) && force <= 0 ) {
+			if( RET_WAS_ERROR(r) ) {
 				aptmethod_shutdown(run);
 				return result;
 			}
 			continue;
 		}
-		r = searchformissing(dbdir,target,force);
-		if( r == RET_NOTHING ) {
+		r = searchformissing(dbdir,target);
+		if( RET_WAS_ERROR(r) )
+			target->incomplete = TRUE;
+		else if( r == RET_NOTHING ) {
 			(void)downloadcache_free(cache);
 			continue;
 		}
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			(void)downloadcache_free(cache);
 			aptmethod_shutdown(run);
 			return result;
 		}
-		r = upgradelist_enqueue(target->upgradelist,cache,filesdb,force);
+		r = upgradelist_enqueue(target->upgradelist,cache,filesdb);
+		if( RET_WAS_ERROR(r) )
+			target->incomplete = TRUE;
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			(void)downloadcache_free(cache);
 			aptmethod_shutdown(run);
 			return result;
@@ -1803,7 +1928,7 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 			fprintf(stderr,"Getting packages for %s's %s...\n",d->distribution->codename,target->target->identifier);
 		r = aptmethod_download(run,methoddir,filesdb);
 		RET_UPDATE(result,r);
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			(void)downloadcache_free(cache);
 			aptmethod_shutdown(run);
 			return result;
@@ -1813,14 +1938,17 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 
 		if( verbose >= 0 )
 			fprintf(stderr,"Installing/removing packages for %s's %s...\n",d->distribution->codename,target->target->identifier);
-		r = upgradelist_install(target->upgradelist,dbdir,filesdb,refs,force,target->ignoredelete,dereferencedfilekeys);
+		r = upgradelist_install(target->upgradelist,dbdir,filesdb,refs,target->ignoredelete,dereferencedfilekeys);
+		if( RET_WAS_ERROR(r) )
+			target->incomplete = TRUE;
 		RET_UPDATE(result,r);
 		upgradelist_free(target->upgradelist);
 		target->upgradelist = NULL;
-		if( RET_WAS_ERROR(r) && force <= 0 ) {
+		if( RET_WAS_ERROR(r) ) {
 			aptmethod_shutdown(run);
 			return result;
 		}
+		markdone(target);
 	}
 	if( verbose > 0 )
 		fprintf(stderr,"Shutting down aptmethods...\n");
@@ -1830,13 +1958,17 @@ static retvalue singledistributionupdate(const char *dbdir,const char *methoddir
 	return result;
 }
 
-retvalue updates_iteratedupdate(const char *confdir,const char *dbdir,const char *distdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *distributions,int force,bool_t nolistsdownload,bool_t skipold,struct strlist *dereferencedfilekeys) {
+retvalue updates_iteratedupdate(const char *confdir,const char *dbdir,const char *distdir,const char *methoddir,filesdb filesdb,references refs,struct update_distribution *distributions,bool_t nolistsdownload,bool_t skipold,struct strlist *dereferencedfilekeys, enum exportwhen export) {
 	retvalue result,r;
 	struct update_distribution *d;
 
 	if( nolistsdownload ) {
 		if( verbose >= 0 )
 			fprintf(stderr,"Warning: As --nolistsdownload is given, index files are NOT checked.\n");
+		if( skipold && verbose >= 0 ) {
+			fprintf(stderr,"Ignoring --skipold because of --nolistsdownload\n");
+		}
+		skipold = FALSE;
 	}
 
 	result = RET_NOTHING;
@@ -1847,12 +1979,14 @@ retvalue updates_iteratedupdate(const char *confdir,const char *dbdir,const char
 			if( verbose >= 0 )
 				fprintf(stderr,"Warning: Override-Files of '%s' ignored as not yet supported while updating!\n",d->distribution->codename);
 		}
-		r = singledistributionupdate(dbdir,methoddir,filesdb,refs,d,force,nolistsdownload,skipold,dereferencedfilekeys);
-		RET_UPDATE(result,r);
-		if( RET_IS_OK(r) ) {
-			r = distribution_export(d->distribution,confdir,dbdir,distdir,TRUE);
-			RET_UPDATE(result,r);
+		if( d->distribution->tracking != dt_NONE ) {
+			fprintf(stderr,"WARNING: Updating does not update trackingdata. Trackingdata of %s will be outdated!\n",d->distribution->codename);
 		}
+		r = singledistributionupdate(dbdir,methoddir,filesdb,refs,d,nolistsdownload,skipold,dereferencedfilekeys);
+		RET_ENDUPDATE(d->distribution->status,r);
+		RET_UPDATE(result,r);
+		r = distribution_export(export,d->distribution,confdir,dbdir,distdir,filesdb);
+		RET_UPDATE(result,r);
 	}
 	return result;
 }
