@@ -1,9 +1,8 @@
 /*  This file is part of "reprepro"
- *  Copyright (C) 2004,2005 Bernhard R. Link
+ *  Copyright (C) 2004,2005,2006 Bernhard R. Link
  *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
+ *  it under the terms of the GNU General Public License version 2 as 
+ *  published by the Free Software Foundation.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -158,7 +157,7 @@ retvalue upgradelist_initialize(struct upgradelist **ul,struct target *t,const c
 	}
 
 	/* Beginn with the packages currently in the archive */
-	r = packages_foreach(t->packages,save_package_version,upgrade,0);
+	r = packages_foreach(t->packages,save_package_version,upgrade);
 	r2 = target_closepackagesdb(t);
 	RET_UPDATE(r,r2);
 
@@ -290,7 +289,7 @@ static retvalue upgradelist_trypackage(void *data,const char *chunk){
 			free(version);
 			return RET_ERROR_OOM;
 		}
-		assert(upgrade->currentaptmethod!=NULL);
+//		assert(upgrade->currentaptmethod!=NULL);
 		new->deleted = FALSE; //to be sure...
 		new->aptmethod = upgrade->currentaptmethod;
 		new->name = packagename;
@@ -366,9 +365,14 @@ static retvalue upgradelist_trypackage(void *data,const char *chunk){
 		if( versioncmp == 0 ) {
 		/* we are replacing a package with the same version,
 		 * so we keep the old one for sake of speed. */
-		// TODO: add switch to force reimport of everything...
+			if( current->deleted && 
+				current->version != current->new_version) {
+				/* remember the version for checkupdate/pull */
+				free(current->new_version);
+				current->new_version = version;
+			} else
+					free(version);
 			current->deleted = FALSE;
-			free(version);
 			free(packagename);
 			return RET_NOTHING;
 		}
@@ -381,9 +385,13 @@ static retvalue upgradelist_trypackage(void *data,const char *chunk){
 			int vcmp = 1;
 			(void)dpkgversions_cmp(version,current->version_in_use,&vcmp);
 			if( vcmp == 0 ) {
-				current->deleted = FALSE;
 				current->version = current->version_in_use;
-				free(version);
+				if( current->deleted ) {
+					free(current->new_version);
+					current->new_version = version;
+				} else
+					free(version);
+				current->deleted = FALSE;
 				free(packagename);
 				return RET_NOTHING;
 			}
@@ -404,7 +412,7 @@ static retvalue upgradelist_trypackage(void *data,const char *chunk){
 		free(current->new_version);
 		current->new_version = version;
 		current->version = version;
-		assert(upgrade->currentaptmethod!=NULL);
+//		assert(upgrade->currentaptmethod!=NULL);
 		current->aptmethod = upgrade->currentaptmethod;
 		strlist_move(&current->new_filekeys,&files);
 		strlist_move(&current->new_md5sums,&md5sums);
@@ -415,14 +423,36 @@ static retvalue upgradelist_trypackage(void *data,const char *chunk){
 	return RET_OK;
 }
 
-retvalue upgradelist_update(struct upgradelist *upgrade,struct aptmethod *method,const char *filename,upgrade_decide_function *decide,void *decide_data,int force){
+retvalue upgradelist_update(struct upgradelist *upgrade,struct aptmethod *method,const char *filename,upgrade_decide_function *decide,void *decide_data){
 
 	upgrade->last = NULL;
 	upgrade->currentaptmethod = method;
 	upgrade->predecide = decide;
 	upgrade->predecide_data = decide_data;
 
-	return chunk_foreach(filename,upgradelist_trypackage,upgrade,force,FALSE);
+	return chunk_foreach(filename,upgradelist_trypackage,upgrade,FALSE);
+}
+
+static retvalue try(void *data,UNUSED(const char *package),const char *chunk) {
+	return upgradelist_trypackage(data,chunk);
+}
+
+
+retvalue upgradelist_pull(struct upgradelist *upgrade,struct target *source,upgrade_decide_function *predecide,void *decide_data,const char *dbdir) {
+	retvalue result,r;
+
+	upgrade->last = NULL;
+	upgrade->currentaptmethod = NULL;
+	upgrade->predecide = predecide;
+	upgrade->predecide_data = decide_data;
+
+	r =  target_initpackagesdb(source,dbdir);
+	if( RET_WAS_ERROR(r) )
+		return r;
+	result = packages_foreach(source->packages,try,upgrade);
+	r = target_closepackagesdb(source);
+	RET_UPDATE(result,r);
+	return result;
 }
 
 /* mark all packages as deleted, so they will vanis unless readded or reholded */
@@ -455,7 +485,7 @@ retvalue upgradelist_listmissing(struct upgradelist *upgrade,filesdb files){
 }
 
 /* request all wanted files in the downloadlists given before */
-retvalue upgradelist_enqueue(struct upgradelist *upgrade,struct downloadcache *cache,filesdb filesdb,int force) {
+retvalue upgradelist_enqueue(struct upgradelist *upgrade,struct downloadcache *cache,filesdb filesdb) {
 	struct package_data *pkg;
 	retvalue result,r;
 	result = RET_NOTHING;
@@ -468,14 +498,41 @@ retvalue upgradelist_enqueue(struct upgradelist *upgrade,struct downloadcache *c
 				&pkg->new_filekeys,
 				&pkg->new_md5sums);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				break;
 		}
 	}
 	return result;
 }
 
-retvalue upgradelist_install(struct upgradelist *upgrade,const char *dbdir,filesdb files,references refs,int force, bool_t ignoredelete, struct strlist *dereferencedfilekeys){
+/* delete all packages that will not be kept (i.e. either deleted or upgraded) */
+retvalue upgradelist_predelete(struct upgradelist *upgrade,const char *dbdir,references refs,struct strlist *dereferencedfilekeys) {
+	struct package_data *pkg;
+	retvalue result,r;
+	result = RET_NOTHING;
+	assert(upgrade != NULL);
+
+	result = target_initpackagesdb(upgrade->target,dbdir);
+	if( RET_WAS_ERROR(result) )
+		return result;
+	for( pkg = upgrade->list ; pkg != NULL ; pkg = pkg->next ) {
+		if( pkg->version_in_use != NULL &&
+				(pkg->version == pkg->new_version || pkg->deleted)) {
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
+				r = target_removepackage(upgrade->target,refs,pkg->name,dereferencedfilekeys,NULL);
+			RET_UPDATE(result,r);
+			if( RET_WAS_ERROR(r))
+				break;
+		}
+	}
+	r = target_closepackagesdb(upgrade->target);
+	RET_ENDUPDATE(result,r);
+	return result;
+}
+
+retvalue upgradelist_install(struct upgradelist *upgrade,const char *dbdir,filesdb files,references refs, bool_t ignoredelete, struct strlist *dereferencedfilekeys){
 	struct package_data *pkg;
 	retvalue result,r;
 
@@ -485,24 +542,35 @@ retvalue upgradelist_install(struct upgradelist *upgrade,const char *dbdir,files
 	result = target_initpackagesdb(upgrade->target,dbdir);
 	if( RET_WAS_ERROR(result) )
 		return result;
+	result = RET_NOTHING;
 	for( pkg = upgrade->list ; pkg != NULL ; pkg = pkg->next ) {
 		if( pkg->version == pkg->new_version && !pkg->deleted ) {
 			r = files_expectfiles(files,&pkg->new_filekeys,
 					&pkg->new_md5sums);
-			if( ! RET_WAS_ERROR(r) )
+			if( ! RET_WAS_ERROR(r) ) {
 				/* upgrade (or possibly downgrade) */
 // TODO: trackingdata?
-				r = target_addpackage(upgrade->target,refs,
-				pkg->name,pkg->new_version,pkg->new_control,
-				&pkg->new_filekeys,TRUE,dereferencedfilekeys,NULL,0);
+				if( interupted() )
+					r = RET_ERROR_INTERUPTED;
+				else
+					r = target_addpackage(upgrade->target,
+						refs, pkg->name,
+						pkg->new_version,
+						pkg->new_control,
+						&pkg->new_filekeys, TRUE,
+						dereferencedfilekeys, NULL, 0);
+			}
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				break;
 		}
 		if( pkg->deleted && pkg->version_in_use != NULL && !ignoredelete ) {
-			r = target_removepackage(upgrade->target,refs,pkg->name,dereferencedfilekeys,NULL);
+			if( interupted() )
+				r = RET_ERROR_INTERUPTED;
+			else
+				r = target_removepackage(upgrade->target,refs,pkg->name,dereferencedfilekeys,NULL);
 			RET_UPDATE(result,r);
-			if( RET_WAS_ERROR(r) && force <= 0 )
+			if( RET_WAS_ERROR(r) )
 				break;
 		}
 	}
@@ -513,8 +581,12 @@ retvalue upgradelist_install(struct upgradelist *upgrade,const char *dbdir,files
 
 void upgradelist_dump(struct upgradelist *upgrade){
 	struct package_data *pkg;
+	
+	assert(upgrade != NULL);
 
 	for( pkg = upgrade->list ; pkg != NULL ; pkg = pkg->next ) {
+		if( interupted() )
+			return;
 		if( pkg->deleted ) {
 			if( pkg->version_in_use != NULL && 
 					pkg->new_version != NULL ) {
